*-- File: HEBREW.CC

CLASS Hebrew
*----------------------------------------------------------------------
*-- Description:
*
*      This class contains methods of working with dates aimed 
*      specifically at the Hebrew calendar. All of the code here was 
*      written by Jay Parsons.
*
*-- Programmers: 
*
*      Ken Mayer
*      Jay Parsons (CIS: 72662,1302)
*
*-- History: 
*
*      11/25/1996 -- original version of these routines as a class.
*
*-- Instantiation: 
*
*      set procedure to hebrew.cc additive
*      _app.Hebrew = new Hebrew()
*
*-- Methods:
*      More details on each of these is contained in the commentary for
*      each method ...
*
*--    Release()
*        Releases this object from memory
*
*--    CivilDate()
*        Hebrew Calendar date to civil calendar date.
*
*--    Dat2Heb()
*        Converts dBASE date to date on Hebrew calendar
*
*--    Dechiyoth()
*        Adjusts date of Rosh Hoshana for the four dechiyoth
*
*--    HebrewDate()
*        Converts a date to corresponding date on Hebrew Calendar
*
*--    IsLeapH()
*        Hebrew Calendar Leap Year
*
*--    Kebiah()
*        Returns the Kebiah of a year
*
*--    NormalH()
*        Normalize date and numbers of hours and chalokim
*
*--    Roshashana()
*        Returns date of Rosh Hashanah of a given Hebrew year.
*
*-- Properties:
*
*      None
*----------------------------------------------------------------------

   FUNCTION Release
   *-------------------------------------------------------------------
   *-- Programmer..: Ken Mayer
   *-- Date........: 11/24/1996
   *-- Notes.......: Used to release this object from memory.
   *-- Written for.: Visual dBASE 5.5
   *-- Rev. History: 11/24/1996 -- Original Code
   *-- Calls.......: none
   *-- Usage.......: hebrew.Release()
   *-- Example.....: _app.Hebrew.Release()
   *--               _app.Hebrew = NULL   && not required, but a good 
   *--                                    && idea
   *-- Returns.....: Null
   *-- Parameters..: None
   *-------------------------------------------------------------------

      release object this

   RETURN NULL
   *-- EoM: Release()

   FUNCTION CivilDate( cDate )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (Compuserve 72662,1302)
   *-- Date........: 08/02/1993
   *-- Notes.......: Converts a Hebrew calendar date to corresponding 
   *--               date on the civil calendar.  The date returned is 
   *--               the civil date that matches the daylight hours of 
   *--               the given Hebrew date; the Hebrew day actually 
   *--               starts the evening before.
   *--                     Supplying the function with a nonexistent 
   *--               30th day of Heshvan or Kishlev is fairly harmless; 
   *--               it will return the civil date of the first of the 
   *--               following month, the correct date to celebrate a 
   *--               birthday falling on the date that does not exist 
   *--               in the given year.  This is **NOT** sufficient to 
   *--               avoid problems with Adar, or with Yahrzeit for 
   *--               deaths occurring on Heshvan 30 or Kishlev 30; 
   *--               users are cautioned against finding anniversaries
   *--               in those cases by supplying the Hebrew month and 
   *--               date of the event and a different Hebrew year as
   *--               parameters to this function.
   *--                      In general, reconverting the civil date
   *--               returned to a Hebrew date and comparing it to the
   *--               original Hebrew date furnished to this function 
   *--               will disclose any possible problems, as the two 
   *--               Hebrew dates will not match.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/27/1993 - Original Version.
   *--               08/02/1993 - Jay Parsons - nX and nY made private,
   *--               method of specifying months changed, error returns
   *--               added.
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: hebrew.Kebiah() 
   *--               hebrew.Roshashana()
   *-- Usage.......: hebrew.Civildate(<cDate> )
   *-- Example.....: ? _app.Hebrew.Civildate( "05/07/5753" )
   *-- Returns.....: dBase date, the corresponding civil date, or {} if 
   *--               the argument fails to specify an existing Hebrew 
   *--               date.
   *-- Parameters..: cDate      = character string holding month, day 
   *--                            and year of Hebrew date, MM/DD/YYYY.  
   *--                            Month must be in range 00-12.  Months
   *--                            01-12 correspond to month order in a 
   *--                            common year, whether or not the year 
   *--                            is common or a leap year, so that 
   *--                            "12" means Ellul in any year.  "00" 
   *--                            means a date in First Adar and is 
   *--                            meaningless in a common year.
   *-------------------------------------------------------------------
      
      local nMo, nDay, nYear, dDate, cKebiah, aDays, nX, nY, dRet
      
      * blank date for error returns
      dRet = {}
      nYear = val( right( cDate, 4 ) )
      if nYear < 1
         RETURN dRet
      endif
      nMo = val( left( cDate, 2 ) )
      if nMo < 0 .or. nMo > 12
         RETURN dRet
      endif
      nDay = val( substr( cDate, 4, 2 ) )
      if nDay < 1 .or. nDay > 30
         RETURN dRet
      endif
      * find the last day of the Hebrew year preceding this one
      dDate = CLASS::Roshashana( nYear ) - 1
      * also Rosh Hashanah of this year and its Kebiah
      cKebiah = CLASS::Kebiah( dDate + 1, ;
                CLASS::Roshashana( nYear + 1 ) )
      * set up array of the days in the months, using Kebiah to find
      * number of days in Heshvan and Kislev, months 2 and 3.  This
      * array is not really required to add the days of each month
      * before the needed one to a running total, but it aids
      * understanding.
      aDays = new Array( 13 )
      aDays[ 1 ] = 30
      aDays[ 2 ] = iif( cKebiah $ "EFGLMN", 30, 29 )
      aDays[ 3 ] = iif( cKebiah $ "ABHIJ", 29, 30 )
      aDays[ 4 ] = 29
      aDays[ 5 ] = 30
      * in leap years, add days of two Adars, change specification of
      * First Adar to month 6 and add one to later months, converting
      * to their actual positions in the leap year.  Otherwise, add
      * 29 days of Adar.
      if cKebiah > "G"
         aDays[ 6 ] = 30
         aDays[ 7 ] = 29
         nX = 7
         do case
            case nMo = 0
                 nMo = 6
            case nMo > 5
                 nMo = nMo + 1
         endcase
      else
         if nMo = 0
            RETURN dRet
         endif
         aDays[ 6 ] = 29
         nX = 6
      endif
      * for the remaining months, alternate 29 and 30 days
      nY = 1
      do while nY < 7
         aDays[ nX + nY ] = 29 + mod( nY, 2 )
         nY = nY + 1
      enddo
      * add up the days in months gone by
      nX = 1
      do while nX < nMo
         dDate = dDate + aDays[ nX ]
         nX = nX + 1
      enddo
      * and the specified days in this month, if they exist
      if nDay <= aDays[ nMo ]
         dRet = dDate + nDay
      endif

   RETURN dRet
   *-- EoF: CivilDate()

   FUNCTION Dat2Heb( dDate )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (Compuserve 72662,1302)
   *-- Date........: 08/03/1993
   *-- Notes.......: Converts a date to corresponding date on the 
   *--               Hebrew calendar.  The date returned is the Hebrew 
   *--               date that matches the daylight hours of the given 
   *--               civil date; the Hebrew date actually starts the 
   *--               evening before. Do not use this to convert dates 
   *--               before 1582 C.E., because the dBASE specification 
   *--               of such dates is erroneous.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/27/1993 - Original version.
   *--               08/03/1993 - Jay Parsons - conformed to change in
   *--               method of specifying months
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: hebrew.Kebiah()
   *--               hebrew.Roshashana() 
   *-- Usage.......: hebrew.Dat2Heb( <dDate> )
   *-- Example.....: ? _app.Hebrew.Dat2Heb( {03/27/93} )
   *-- Returns.....: a character string giving the month, day and year 
   *--               of the date in numerals, such as "07/05/5753". The
   *--               month will be given as the number of that month in 
   *--               a common year, whether or not the year is a leap 
   *--               year, so that "07" always means Nisan.  A date in 
   *--               First Adar in a leap year will be returned as 
   *--               month "00".
   *-- Parameters..: dDate      = a dBASE civil date to convert
   *-------------------------------------------------------------------
      
      local nYear, dFirst, dSecond, cKebiah, aDays, nX, nY, nDay, cD
      
      * convert the year to Hebrew year that begins in the civil year
      nYear = year( dDate ) + 3761
      * find Rosh Hashanah in the year
      dFirst = CLASS::Roshashana( nYear )
      * back up if date is in the preceding Hebrew year
      if dFirst > dDate
         nYear = nYear - 1
         dSecond = dFirst
         dFirst = CLASS::Roshashana( nYear )
      else
         dSecond = CLASS::Roshashana( nYear + 1 )
      endif
      cD = "/" + str( nYear, 4 )
      * find number of the day within the Hebrew year, and the Kebiah
      nDay = 1 + dDate - dFirst
      cKebiah = CLASS::Kebiah( dFirst, dSecond )
      * set up array of the days in the months, using Kebiah to find
      * number of days in Heshvan and Kislev, months 2 and 3.  This
      * array is not really required to add the days of each month
      * to a running total, but it aids understanding.
      aDays = new Array( 13 )
      aDays[ 1 ] = 30
      aDays[ 2 ] = iif( cKebiah $ "EFGLMN", 30, 29 )
      aDays[ 3 ] = iif( cKebiah $ "ABHIJ", 29, 30 )
      aDays[ 4 ] = 29
      aDays[ 5 ] = 30
      * in leap years, add days of two Adars, change specification of
      * First Adar to month 6 and add one to later months, converting
      * to their actual positions in the leap year.  Otherwise, add
      * 29 days of Adar.
      if cKebiah > "G"
         aDays[ 6 ] = 30
         aDays[ 7 ] = 29
         nX = 7
      else
         aDays[ 6 ] = 29
         nX = 6
      endif
      * for the remaining months, alternate 29 and 30 days
      nY = 1
      do while nY < 7
         aDays[ nX + nY ] = 29 + mod( nY, 2 )
         nY = nY + 1
      enddo
      * reduce the day of the year by the days in months past
      nX = 1
      do while nDay > aDays[ nX ]
         nDay = nDay - aDays[ nX ]
         nX = nX + 1
      enddo
      * adjust month representation in leap years
      if cKebiah > "G" .and. nX > 5
         nX = iif( nX = 6, 0, nX - 1 )
      endif
      cD = transform( nX, "@L 99" ) + "/" ;
              + transform( nDay, "@L 99" ) + cD

   RETURN cD
   *-- EoF: Dat2Heb()

   FUNCTION Dechiyoth( dDate, nHrs, nChalokim )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 03/26/1993
   *-- Notes.......: Adjusts date of Rosh Hashanah for the four 
   *--               dechiyoth ( postponements ) required to regularize
   *--               the calendar and prevent either two days in a row 
   *--               of Sabbath at Yom Kippur or Hoshanah Rabbah 
   *--               falling on the Sabbath. Days, hours and chalokim  
   *--               are measured from 6 p.m. Friday evening, 
   *--               considered the beginning of the Sabbath for 
   *--               calendar purposes.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/26/1993
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: hebrew.IsLeapH
   *-- Called by...: hebrew.Roshashana(), primarily
   *-- Usage.......: hebrew.Dechiyoth( <dDate>, <nHrs>, <nChalokim> )
   *-- Example.....: nDay = _app.Hebrew.Dechiyoth( {10/02/1978},11,614 )
   *-- Returns.....: date       = civil date of Rosh Hashanah
   *-- Parameters..: dDate      = civil date of Molad Tishri
   *--               nHrs       = hour of Molad Tishri past 6 pm
   *--               nChalokim  = chalokim past the hour of Molad 
   *--                            Tishri
   *-------------------------------------------------------------------      

      local nMoldow, nFirst, nNidcheh
      
      nMoldow = mod( dow( dDate ), 7 )
      nNidcheh = 0
      nFirst = nMoldow
      * 1) if Molad Tishri is after noon, new moon could not be seen
      *    anywhere that day, so Rosh Hashanah is pushed off (nidcheh)
      *    to the next day--it is to be celebrated on the day the new
      *    moon could first be seen.
      if nHrs * 1080 + nChalokim > 18 * 1080
         nNidcheh = 1
         nFirst = nFirst + 1
      endif
      * 2) if Rosh Hashanah would be Sunday, Wednesday or Friday, it
      *    is nidcheh ( again if dechiyah 1 applied ) so that Yom
      *    Kippur will not fall on Friday or Sunday creating two days
      *    of Sabbath in a row, nor Hoshanah Rabbah fall on Sabbath.
      if nFirst = 1 .or. nFirst = 4 .or. nFirst = 6
         nNidcheh = nNidcheh + 1
      endif
      * 3) if Molad Tishri in a common year is Tuesday and the next
      *    Molad Tishri would be after noon of the Sabbath, this Rosh
      *    Hashanah is nidcheh twice so the next Rosh Hashanah will
      *    not have to be nidcheh--the result would be that this year
      *    would have 356 days, too many.
      if .not. CLASS::IsLeapH( year( dDate ) + 3761 ) .and. ;
          nMoldow  = 3 ;
         .and. 1080 * nHrs + nChalokim > 9 * 1080 + 204   && 9 hr 204
         nNidcheh = 2
      endif
      * 4) if Molad Tishri in a common year following a leap year
      *    occurs Monday morning so late that the preceding leap year
      *    must have started on Thursday, this Rosh Hashanah is
      *    nidcheh so the preceding leap year will not be left with
      *    382 days, too few.
      if CLASS::IsLeapH( year( dDate ) + 3760 ) .and. nMoldow = 2 ;
         .and. 1080 * nHrs + nChalokim > 15 * 1080 + 589 && 15 hr 589
         nNidcheh = 1
      endif

   RETURN nNidcheh
   *-- EoF: Dechiyoth()

   FUNCTION HebrewDate( dDate )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (Compuserve 72662,1302)
   *-- Date........: 07/03/1993
   *-- Notes.......: Converts a date to corresponding date on the 
   *--               Hebrew calendar.  The date returned is the Hebrew 
   *--               date that matches the daylight hours of the given 
   *--               civil date; the Hebrew date actually starts the 
   *--               evening before.
   *--                     Of this family of functions, this is the 
   *--               only one containing the names of the months, to 
   *--               simplify changing the spelling of any of them. The 
   *--               ruler may have to be adjusted if the names are 
   *--               changed, as all names must be the same length 
   *--               including padding.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/27/1993 - Jay Parsons - original function
   *--               07/03/1993 - Jay Parsons - nMolen variable added 
   *--               to simplify changing names of months
   *--               08/03/1993 - Jay Parsons - revised to conform to
   *--               changes to Dat2Heb() in specification of months.
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: hebrew.Dat2Heb() 
   *--               hebrew.IsLeapH() 
   *-- Usage.......: hebrew.HebrewDate( <dDate> )
   *-- Example.....: ? _app.Hebrew.HebrewDate( {03/27/93} )
   *-- Returns.....: a character string giving the month, day and year 
   *--               of the date, such as "Nisan 5, 5753"
   *-- Parameters..: dDate      = a dBASE date
   *-------------------------------------------------------------------
      
      local cDate, cMonths, nMo, cMo, nDay, cDay, nYear, nMolen, nM1
      
      * ruler-> 012345678901234567890123456789012345678901
      cMonths = "Tishri HeshvanKislev Tebeth Shebat Adar   " + ;
                "Nisan  Iyar   Sivan  Tammuz Ab     Ellul  "
      cDate = CLASS::Dat2Heb( dDate )
      nYear = val( right( cDate, 4 ) )
      nMo = val( left( cDate, 2 ) )
      nDay = val( substr( cDate, 4, 2 ) )
      nMolen = len( cMonths ) / 12
      nM1 = nMo + iif( nMo = 0, 6, 0 ) - 1
      cMo = trim( substr( cMonths, nMolen * nM1 + 1, ;
                             nMolen ) ) ;
               + " " + ltrim( str( nDay, 2 ) ) + ", " ;
               + right( cDate, 4 )
      do case
         case nMo = 0
              cMo = "First "+ cMo
         case nMo = 6 .and. CLASS::isLeapH( nYear )
              cMo = "Second " + cMo
      endcase

   RETURN cMo
   *-- EoF: HebrewDate()

   FUNCTION IsLeapH( nYear )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 03/27/1993
   *-- Notes.......: Returns .T. or .F. whether a Hebrew year is a leap
   *--               year.  The formula is a hack; it returns .T. when 
   *--               the position of the year in the cycle is 3, 6, 8, 
   *--               11, 14, 17 or 19 ( the last, because its modulus 
   *--               19 is 0. )
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/27/1993 -- Original
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: None
   *-- Usage.......: hebrew.IsLeapH( <nYear> )
   *-- Example.....: ? _app.Hebrew.IsLeapH( 1993 + 3761 )
   *-- Returns.....: logical    = true if the year is a leap year, or 
   *--                            false
   *-- Parameters..: nYear      = number of the year on the Hebrew 
   *--                            calendar
   *-------------------------------------------------------------------
      
      local nY
      
      nY = mod( nYear, 19 )
      
   RETURN mod( nY + int( ( nY + 3 ) / 11 ), 3 ) = 0
   *-- EoF: IsLeapH()
   
   FUNCTION Kebiah( dRosh1, dRosh2 )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (Compuserve 72662,1302)
   *-- Date........: 03/26/1993
   *-- Notes.......: Kebiah of a year, using Arthur Spier's notation.
   *--                     This tells whether the year in the Hebrew
   *--               calendar is defective, regular or excessive and
   *--               whether or not a leap year.  In order to limit the
   *--               recalculations of Rosh Hashanah in date 
   *--               conversions that would be needed if this function 
   *--               were called with the year only, it requires the 
   *--               civil dates of the two Rosh Hashanahs at start and 
   *--               end as parameters.
   *--                     Error checking is limited to determining 
   *--               whether the days of the week of the two dates 
   *--               given and number of days between them are 
   *--               possible; no check is made that either is the 
   *--               correct date of Rosh Hashanah.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/26/1993 -- Original
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: None
   *-- Usage.......: hebrew.Kebiah(<dRosh1>,dRosh2>)
   *-- Example.....: c = _app.Hebrew.Kebiah( {09/16/1993},{09/06/1994} )
   *-- Returns.....: a letter from A through N, signifying as follows:
   *--                     A - G common year of 12 months
   *--                     H - N leap year of 13 months
   *--                 First day of      length and type     First day
   *--          Letter Rosh Hashanah        of year         of Passover
   *--            A     Monday        353 days, defective      Tuesday
   *--            B     Sabbath       353 days, defective      Sunday
   *--            C     Tuesday       354 days, regular        Thursday
   *--            D     Thursday      354 days, regular        Sabbath
   *--            E     Monday        355 days, excessive      Thursday
   *--            F     Thursday      355 days, excessive      Sunday
   *--            G     Sabbath       355 days, excessive      Tuesday
   *--            H     Monday        383 days, defective      Thursday
   *--            I     Thursday      383 days, defective      Sunday
   *--            J     Sabbath       383 days, defective      Tuesday
   *--            K     Tuesday       384 days, regular        Sabbath
   *--            L     Monday        385 days, excessive      Sabbath
   *--            M     Thursday      385 days, excessive      Tuesday
   *--            N     Sabbath       385 days, excessive      Thursday
   *--            or "" signifying error in dates passed to function.
   *-- Parameters..: dRosh1      = civil date of Rosh Hashanah that 
   *--                             starts the year
   *--               dRosh2      = civil date of Rosh Hashanah starting 
   *--                             the following year
   *-------------------------------------------------------------------
      
      local dR1, dR2, nDays, nDow, cRet
      
      dR1 = min( dRosh1, dRosh2 )
      dR2 = max( dRosh1, dRosh2 )
      nDays = dR2 - dR1
      nDow = dow( dR2 )
      * Rosh Hashanah cannot be on Sunday, Wednesday or Friday
      if nDow = 1 .or. nDow = 4 .or. nDow = 6
         RETURN ""
      endif
      cRet = space( 7 )     && will return "" if wrong # of days
      do case
         * Days       SMTWHFS   && if the char in cRet below the dow
         * Dows       1234567   && is blank, can't be Rosh Hashanah
         case nDays = 353
              cRet = " A    B"
         case nDays = 354
              cRet = "  C D  "
         case nDays = 355
              cRet = " E  F G"
         case nDays = 383
              cRet = " H  I J"
         case nDays = 384
              cRet = "  K    "
         case nDays = 385
              cRet = " L  M N"
      endcase
   
   RETURN trim( substr( cRet, dow( dR1 ), 1 ) )
   *-- EoF: Kebiah()

   FUNCTION NormalH( dDate, nHrs, nChalokim )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 03/26/1993
   *-- Notes.......: Normalize date and numbers of hours and chalokim
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/26/1993 -- Original
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: None
   *-- Called by...: Any
   *-- Usage.......: hebrew.NormalH(<dDate>, <nHrs>, <nChalokim>)
   *-- Example.....: _app.Hebrew.Normalh({03/13/1993}, 39, 1452)
   *-- Returns.....: None
   *-- Parameters..: dDate         = a dBASE civil date
   *--               nHrs          = number of hours
   *--               nChalokim     = number of chalokim
   *-- Side Effects: This procedure changes its parameters.
   *-------------------------------------------------------------------
      
      nHrs       = nHrs + floor( nChalokim / 1080 )
      nChalokim  = mod( nChalokim, 1080 )
      dDate      = dDate + floor( nHrs / 24 )
      nHrs       = mod( nHrs, 24 )

   RETURN ""
   *-- EoF: NormalH

   FUNCTION Roshashana( nYear )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (Compuserve 72662,1302)
   *-- Date........: 04/22/1994
   *-- Notes.......: Returns date of Rosh Hashanah of a given Hebrew 
   *--               year. To find date when Rosh Hashanah occurs in a 
   *--               C.E. year, add 3761 to the C.E. Year.  This 
   *--               formidable-looking procedure is actually a 
   *--               straightforward matter of using the date and time 
   *--               of one Molad Tishri, the moment of new moon 
   *--               starting a year, and advancing or setting back 
   *--               that date and time by the number of days, hours 
   *--               and chalokim in each period of 19 years, then by
   *--               the number in each month within the balance of the
   *--               years.  Since we are concerned only with the 
   *--               molads, we need not worry about the number of days
   *--               in the calendar months, but only with the lunar 
   *--               period of 29 days, 12 hours and 793 chalokim and 
   *--               the number of months, giving effect to leap years.
   *--                     A chelek, plural chalokim, is 1/1080 of an 
   *--               hour, or 3 1/3 seconds, and is traditionally used 
   *--               for these calculations, in part because it avoids 
   *--               the need to deal with fractions.
   *--                     Finally, having determined the civil date of 
   *--               the Molad Tishri of the proper year and the hour 
   *--               and chalokim within the date, we call a separate 
   *--               function to determine the applicability of the 
   *--               dechiyoth or postponements that may cause Rosh 
   *--               Hashanah to be pushed to a later date.
   *-- Written for.: dBASE IV, 2.0
   *-- Rev. History: 03/26/1993 - Original Version.
   *--               08/02/1993 - Jay Parsons - nCycles made private.
   *--               04/22/1994 - Jay Parsons - independent of date 
   *--               set().
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/03/1996 -- Minor modifications to work with 
   *--               this class.
   *-- Calls.......: hebrew.NormalH
   *--               hebrew.Dechiyoth
   *-- Usage.......: hebrew.Roshashana( <nYear> )
   *-- Example.....: _app.Hebrew.Roshashana( 1993 + 3761 )
   *-- Returns.....: date  = civil date of Rosh Hashanah in the 
   *--                       year
   *-- Parameters..: nYear = number of year in the Hebrew calendar
   *-------------------------------------------------------------------
      
      local cDateSet,dMoldate, nMolhr, nMolch, nYrs, nMoons, nCycles
      
      * we use the Molad of Tishri 5739, October 2, 1978, as the base
      * for no good reason except it started the most recent cycle.
      cDateSet  = set("DATE")
      set date AMERICAN
      dMoldate  = { 10/02/1978 } && The Hebrew date started 6 p.m.
      nMolhr    = 11             && Oct. 1, 1978.  The molad (new
      nMolch    = 614            && moon) was 11 614/1080 hrs later
      set date &cDateSet.
      * adjust for each full 19-year cycle ( machzor koton ), which
      * includes 6939 days, 16 hours and 595 chalokim.
      * if adjusting backward, go to beginning of the cycle in which
      * the specified year falls
      nYrs      = nYear - 5739   && years of difference to adjust.
      nCycles   = floor( nYrs / 19 )
      dMoldate  = dMoldate + 6939 * nCycles
      nMolhr    = nMolhr + 16 * nCycles
      nMolch    = nMolch + 595 * nCycles
      * adjust for leftover months within the cycle; the formula was
      * hacked to return the correct number of months for any number
      * of years from 0 through 19 at the start of a 19-year cycle.
      nYrs      = nYrs - 19 * nCycles
      nMoons    = int( .01 + nYrs * 12.374 )
      * one lunar month is 29 days 12 hours 793 chalokim
      dMoldate  = dMoldate +  29 * nMoons
      nMolhr    = nMolhr   +  12 * nMoons
      nMolch    = nMolch   + 793 * nMoons
      CLASS::NormalH( dMoldate, nMolhr, nMolch )
  
   RETURN dMoldate + CLASS::Dechiyoth( dMolDate, nMolhr, nMolch )
   *-- EoF: Roshashana()

ENDCLASS
*-- End of Class:: Hebrew.cc

*-----------------------------------------------------------------------
*-- End of File: HEBREW.CC
*-----------------------------------------------------------------------