*-- FILE: ROOT.CC

CLASS Root
*-----------------------------------------------------------------------
*-- Programmer..: Jay Parsons (CIS: 72662,1302)
*--		  Adapted from the function ZeroIn()
*--		  by Tony Lima (CIS: 72331,3724) and
*--		  Jay Parsons published 04/13/1992
*-- Date........: 10/20/1996	
*-- Notes.......: Finds a root of a continuous function.
*--
*--               An object of this class has the properties needed to
*--		  identify a function, two starting values of the argument
*--		  to the function, an offset from zero of solutions, and 
*--		  the accuracy of result desired, as well as a flag property
*--		  that may be inspected to find out what went wrong. The sole
*--		  method of the class finds and returns a root of the function,
*--               if it can.

*--	          The method can find a root only if the function it is
*--		  passed is continuous. That is, when y = f(x) is plotted
*--		  using Cartesian coordinates, the plot over all values of 
*--               x is a line, curved or straight, with no breaks.         					

*--		  If the function y has a derivative, do not use this class.
*--		  The existence of a derivative implies that a root may be 
*--		  found directly without the trial-and-error tactics used here.
*--		
*--	          The starting values are two values of the independent 
*--		  variable,  x, that result in values of y with different
*--		  signs (after offset). Only with such starting values is it
*--		  clear that, if the function is continuous, there must be
*		  some intermediate value of x which results in y = 0, after
*		  offset. This value of x is the value we want. 			
*--
*--               The findroot() method is slow, so a maximum number of 
*--               iterations is a property. If the number 
*--               of iterations is exceeded, the function will fail to 
*--               find a root.  if this occurs, pick different original 
*--               "x" values, increase the number of iterations or 
*--               decrease the accuracy required.  Once an approximate 
*--               root is found, you may use values of x close on 
*--               either side and reduce the error allowed to find an
*--               improved solution.
*--             
*--               NOTE ESPECIALLY - There is NO guarantee that a root 
*--               returned by this function is the only one, or the most
*--               meaningful. It depends on the function that this 
*--               function calls, but if that function has several 
*--               roots, any of them may be returned. This can easily 
*--               happen with such called functions as net present value
*--               where the cash flows alternate from positive to 
*--               negative and back, and in many other "real life" 
*--               cases. See the discussion of @IRR in the documentation
*--               of a good spreadsheet program such as Quattro Pro for 
*--               further information.
*--             
*--               The algorithm used by this function is a "secant and 
*--               bisect" search.  The "secant" is the line connecting 
*--               two x,y points on a graph using standard Cartesian 
*--               coordinates. Where the secant crosses the y axis is 
*--               the best guess for the value of x that will have 
*--               y = 0, and will be correct if the function is linear
*--               between the two points.  The basic strategy is to 
*--               calculate y at that value of x, then keep the new x 
*--               and that one of the old x values that had a y-value 
*--               of opposite sign, and reiterate to close in.
*--             
*--               If the function is a simple curve, repeated secants
*--		  will approach, but never reach, an x-value with a 
*--		  y-value equal to 0. In this case the function shifts
*--		  to choosing the new x halfway between the old ones,
*--		  bisecting the difference and always reducing the 
*--		  bracket by half, for a while.
*--             
*--               While this function needs to find y-values different in 
*--		  sign, it may be used to find an x corresponding to any
*--		  nonzero value of y by providing an offset equal to that
*--		  value of y.
*--
*--		  For an example of the use of this function, see Bond.cc
*--		  in the CompuServe DBASEWIN libraries.	
*--		  
*-- Written for.: Visual dBASE 5.5
*-- Rev. History: Original function 1990.
*--             : Modified to take optional parameters, 4/13/1992
*--		: Converted to a class, 09/17/1996
*--		: Error in secant/bisect decision fixed, 10/20/1996
*--		
*-- Properties..: fp  		= pointer to the function to call, required
*--		: x1		= a starting x value, required
*-- 		: x2		= another starting x value, required. It is also
*--				  required that fp(x1) and fp(x2), in each case
*--				  less offset, have different signs, or a flag 4
*--				  error occurs
*--		: offset	= required value of fp(x). If omitted, 0 assumed.
*--		: maxiters      = maximum iterations before quitting. If omitted,
*--				  100 assumed
*--		: flag          = a flag to show result, as follows:
*--  	                	  1 success - root found within error allowed
*--                     	  2 success - root was found exactly
*--                     	  3 error   - f values not converging
*--                     	  4 error   - x1 and x2 badly chosen
*--                               5 error   - maximum iterations exceeded
*--		: abserror      = the absolute error allowed in the result.
*--				  If omitted, .0001 is used. 
*--		: iterations    = count of iterations, for testing
*--
*-- Parameters..: None to instantiate the class or to findroot(). 		
*--		  The function pointed to by the fp property must take
*--		  one formal parameter, the new estimate of x, and 
*--		  return the value of fp(x).
*--               
*-- Returns.....: The findroot() method returns a value of x that
*--		  is a root, but only if the flag property is 1 or 2.	
*-----------------------------------------------------------------------

   this.fp         = .F.
   this.x1         = 0
   this.x2         = 0
   this.offset     = 0
   this.maxiters   = 100
   this.flag       = 0
   this.abserror   = .0001
   this.iterations = 0   

   function findroot
      local nSplits, fBracket, fNearx, fFarx, fFary, fNeary, fMaxAbs
      local fOldx, fOldy, fDiffx, fAbsdiff, fSecant

      *clean out impossible calls
      nSplits = 0
      fNearX = this.x1
      fFarx = this.x2
      fBracket = abs (fFarx - fNearx)

      fFarY = this.fp(fFarX) - this.offset
      
      fNearY= this.fp(fNearx) - this.offset 

      if sign(fNearY) = sign(fFarY)
         this.flag = 4
         RETURN 0
      endif

      * Find greater error from root
      fMaxAbs = max(abs(fNearY), abs(fFarY))
      this.flag = 0

      * Main iteration loop

      do while this.flag = 0

         if abs(fFarY) < abs(fNearY)
  
            * Interchange fNearX and fFarx so that
            * fNearX is closer to a solution--
            * abs(fNearY) <= abs(fFarY)

            fOldX  = fNearX
            fOldY  = fNearY
            fNearX = fFarX
            fNearY = fFarY
            fFarX  = fOldX
            fFarY  = fOldY
         endif
 
         fDiffX = fFarX - fNearX
         fAbsDiff = abs(fDiffX)

         * Test whether interval is too small to continue
 
         if fAbsDiff <= 2 * this.abserror
            if abs(fNearY) > fMaxAbs

               * Yes, but we are out of bounds
 
               this.flag = 3
               RETURN 0
            else
  
               * Yes, and we have a solution!
  
               this.flag = 1
               RETURN (fNearX + fFarX)/2 
            endif
         endif

         * Save the last approximation to x and y

         fOldX = fNearX
         fOldY = fNearY

         * Check if reduction in the size of
         * bracketing interval is satisfactory.
         * if not, bisect until it is.

         nSplits = nSplits + 1
         if nSplits >= 4 
            if 4 * fAbsDiff >= fBracket
               fNearX = fNearX + fDiffX / 2
            else
               nSplits = 0
               fBracket = fAbsDiff / 2
            endif   
         endif

         * Unless bisecting, calculate secant	

         if nSplits < 4
            fSecant = (fNearX - fFarX) * fNearY ;
                                  / (fFarY - fNearY)
   
            * But not less than error allowed
 
            fNearX = fNearX + iif(abs(fSecant) < this.abserror, this.abserror * sign(fDiffX), fSecant)
         endif

         * Evaluate the function at the new approximation

         fNearY= this.fp(fNearX) - this.offset

         * if it's exactly zero, we win!  Run with it
 
         if fNearY = 0
            this.flag = 2
            RETURN fNearX
         endif

         * else adjust iteration count and quit if too
         * many iterations with no solution

         this.iterations = this.iterations + 1
         if this.iterations > this.maxiters
            this.flag = 5
            RETURN 0
         endif

         * And finally keep as the new fFarX that one
         * of the previous approximations, fFarX and
         * fOldx, at which the function has a sign opposite
         * to that at the new approximation, fNearx.

         if sign(fNearY) = sign(fFarY)
            fFarX = fOldX
            fFarY = fOldY
         endif

      enddo

ENDCLASS
*-- End of Class: ROOT

*-----------------------------------------------------------------------
*-- End of File: ROOT.CC
*-----------------------------------------------------------------------       